import os
import time
from nmf import *
from markerFinder import *
from linearSVM import *
from feature_selection import *


def clustering_wrapper(adata, output_filename, rank=None, min_group_size=3, top_n_genes=60, rho_threshold=0.2, marker_finder_rho=0.2, write_files=True):

    if write_files is True:
        os.mkdir(output_filename)
        cd = os.getcwd()

    fold_matrix = adata.varm['pseudobulk_folds']

    # get var names for which 'pca_selected_genes' is True
    highly_variable_genes = adata.var.index[adata.var['correlated_genes'] == True]
    print("using corr genes...")
    # filter the full fold matrix with the highly variable events (currently comes from the file PCAbasedFeatureSelection.py ran before this function)
    fold_matrix_hvg = fold_matrix.loc[highly_variable_genes, :]

    # when rank is not 2 or not provided by the user (that they trust more), then automatically calculate number of clusters for NMF to determine (aka rank)
    # requires the determineNMFRank.py script
    if rank is None:
        est_k = determine_nmf_ranks(fold_matrix_hvg)
        rank = est_k

    # run and time the NMF analysis using the runNMF.py script
    start_time = time.time()
    print(("Running NMF analysis with rank set to " + str(rank) + "..."))
    basis_matrix, nmf_clusters = run_nmf(df=fold_matrix_hvg, rank=rank)
    print("--- %s seconds ---" % (time.time() - start_time))

    # remove clusters with less than 3 (min_groups_size) samples/pseudobulk folds in them
    # Count the number of samples in each cluster
    cluster_counts = nmf_clusters['cluster'].value_counts()

    # Filter clusters that have more than 3 samples
    large_clusters = cluster_counts[cluster_counts > min_group_size].index

    # Keep only samples that belong to the large clusters
    nmf_clusters = nmf_clusters[nmf_clusters['cluster'].isin(large_clusters)]

    fold_matrix = fold_matrix.loc[:, nmf_clusters.index]
    fold_matrix_hvg = fold_matrix_hvg.loc[:, nmf_clusters.index]

    start_time = time.time()
    print("Finding marker genes per cluster...")
    markers_df_og, markers_df, marker_heatmap = marker_finder_wrapper(input_df=fold_matrix.transpose(), groups=nmf_clusters, top_n=top_n_genes, rho_threshold=rho_threshold, marker_finder_rho=marker_finder_rho)
    print("--- %s seconds ---" % (time.time() - start_time))

    adata.uns['clusters_preSVM'] = nmf_clusters
    adata.uns['udon_marker_genes_full_preSVM'] = markers_df_og
    adata.uns['udon_marker_genes_top_n_preSVM'] = markers_df
    adata.uns['marker_heatmap_preSVM'] = marker_heatmap

    # remove the clusters with less than 10 markers
    # Count the number of markers in each cluster
    mf_gene_counts = markers_df['top_cluster'].value_counts()

    # Filter clusters that have more than 3 samples
    robust_clusters = mf_gene_counts.index

    # Keep only samples that belong to the large clusters
    markers_df = markers_df[markers_df['top_cluster'].isin(robust_clusters)]

    # filter the nmf_clusters with robust_clusters too
    nmf_clusters = nmf_clusters[nmf_clusters['cluster'].isin(robust_clusters)]

    start_time = time.time()
    print(("Determined " + str(len(markers_df)) + " unique marker genes across all clusters"))
    fold_matrix_with_mf_genes = fold_matrix.loc[markers_df['marker'], nmf_clusters.index]
    print("--- %s seconds ---" % (time.time() - start_time))

    # Linear SVM/SVC training dataset generation by creating centroids for the NMF clusters

    centroids = generate_train_data(fold_matrix_with_mf_genes=fold_matrix_with_mf_genes, groups=nmf_clusters)  # genes by clusters matrix
    centroids = centroids.dropna(axis=0)
    fold_matrix_with_mf_genes = fold_matrix_with_mf_genes.loc[centroids.index, :]

    start_time = time.time()
    print(("Running Linear SVM on final " + str(np.shape(centroids)[1]) + " clusters with..."))
    final_clusters = classify(train=centroids, fold_matrix_with_mf_genes=fold_matrix_with_mf_genes, groups=nmf_clusters)
    print("--- %s seconds ---" % (time.time() - start_time))

    fold_matrix = fold_matrix.loc[:, final_clusters.index]

    start_time = time.time()
    print("Finding final marker genes per cluster...")
    markers_df_og, markers_df, marker_heatmap = marker_finder_wrapper(input_df=fold_matrix.transpose(),
                                                                      groups=final_clusters, top_n=top_n_genes,
                                                                      rho_threshold=rho_threshold, marker_finder_rho=marker_finder_rho)
    print("--- %s seconds ---" % (time.time() - start_time))

    # clean the marker heatmap file for the final clusters in mf genes
    robust_clusters_final = markers_df['top_cluster'].value_counts(sort=False)
    robust_clusters_final = robust_clusters_final[robust_clusters_final > 15].index

    # Keep only samples that belong to the large clusters
    final_clusters_clean = final_clusters[final_clusters['cluster'].isin(robust_clusters_final)]
    markers_df_clean = markers_df[markers_df['top_cluster'].isin(robust_clusters_final)]

    final_samples = pd.concat([pd.Series('row_clusters-flat'), pd.Series(final_clusters_clean.index)])
    final_rows = pd.concat([pd.Series('column_clusters-flat'), pd.Series(markers_df_clean.index)])
    marker_heatmap = marker_heatmap.loc[final_rows, final_samples]

    adata.uns['udon_clusters'] = final_clusters_clean
    adata.uns['udon_marker_genes_full'] = markers_df_og
    adata.uns['udon_marker_genes_top_n'] = markers_df_clean
    adata.uns['marker_heatmap'] = marker_heatmap

    if write_files is True:
        mf_genes_path = os.path.join(cd, output_filename, "marker_genes_all.txt")
        mf_heatmap_path = os.path.join(cd, output_filename, "marker_heatmap.txt")
        final_clusters_path = os.path.join(cd, output_filename, "udon_clusters.txt")
        markers_df_og.to_csv(mf_genes_path, sep="\t")
        marker_heatmap.to_csv(mf_heatmap_path, sep="\t")
        final_clusters.to_csv(final_clusters_path, sep="\t")

    return adata


def find_udon_clusters(adata, output_filename, species, gtf_file_path=None, fold_threshold=1, samples_differing=3, intercorr_threshold=0.4, corr_n_events=5, pca_corr_threshold=0.4, n_components=30, rank=None, min_group_size=3, delta=0.1, p_val=0.05, min_differential_genes=100, top_n_differential_genes=150):

    print("beginning feature selection...")
    adata = feature_selection_wrapper(adata, species=species, gtf_file_path=gtf_file_path, fold_threshold=fold_threshold, samples_differing=samples_differing, intercorr_threshold=intercorr_threshold, corr_n_events=corr_n_events, pca_corr_threshold=pca_corr_threshold, n_components=n_components)

    print("performing NMF clustering...")
    final_clusters, marker_genes = clustering_wrapper(adata, output_filename=output_filename, rank=rank, min_group_size=min_group_size)

    adata.uns['udon_clusters'] = final_clusters
    adata.uns['udon_marker_genes'] = marker_genes

    return adata



import os
import sys
import pysam
import logging
import pathlib
import argparse
from altanalyze3.utilities.logger import setup_logger
from altanalyze3.utilities.helpers import get_version
from altanalyze3.components.intron_count.main import count_introns
from altanalyze3.components.junction_count.main import count_junctions
from altanalyze3.components.aggregate.main import aggregate
from altanalyze3.utilities.io import (
    get_indexed_references,
    is_bam_indexed
)
from altanalyze3.utilities.constants import (
    IntRetCat,
    MAIN_CRH,
    ChrConverter
)


class ArgsParser():

    def __init__(self, args):
        args = args + [""] if len(args) == 0 else args
        self.args, _ = self.get_parser().parse_known_args(args)
        self.resolve_path(["bam", "ref", "tmp", "output", "juncounts", "intcounts"])
        self.assert_args()
        self.set_args_as_attributes()

    def set_args_as_attributes(self):
        for arg, value in vars(self.args).items():
            setattr(self, arg, value)

    def add_common_arguments(self, parser):
        self.common_arguments = [
            ("--loglevel", "Logging level. Default: info", str, "info", ["fatal", "error", "warning", "info", "debug"]),
            ("--threads", "Number of threads to run in parallel where applicable. Default: 1", int, 1, None),
            ("--cpus", "Number of processes to run in parallel where applicable. Default: 1", int, 1, None),
            ("--tmp", "Temporary files location. Default: tmp", str, "tmp", None),
            ("--output", "Output prefix. Default: results", str, "results", None)
        ]
        for param in self.common_arguments:
            parser.add_argument(
                param[0],
                help=param[1],
                type=param[2],
                default=param[3],
                choices=param[4]
            )

    def get_parser(self):
        """
        Defines arguments for parser. Inlcudes subparsers for
        each component of the tool.
        """
        parent_parser = argparse.ArgumentParser(add_help=False)
        general_parser = argparse.ArgumentParser(description="AltAnalyze3")
        subparsers = general_parser.add_subparsers()
        subparsers.required = True
        # Global parameters for all components of the tool
        general_parser.add_argument(                       
            "--version",
            action="version",
            version=get_version(),
            help="Print current version and exit"
        )

        # Intron count parameters
        intron_parser = subparsers.add_parser(
            "intcount",
            parents=[parent_parser],
            help="Count introns reads"
        )
        intron_parser.set_defaults(func=count_introns)
        intron_parser.add_argument(
            "--bam",
            help="Path to the coordinate-sorted indexed BAM file",
            type=str,
            required=True
        )
        intron_parser.add_argument(
            "--ref",
            help="Path to the gene model reference file. Coordinates are treated as 1-based.",
            type=str,
            required=True
        )
        intron_parser.add_argument(
            "--span",
            help="5' and 3' overlap that read should have over a splice-site to be counted",
            type=int,
            default=10
        )
        intron_parser.add_argument(
            "--strandness",
            help=" ".join(
                [
                    "Strand specificty of the RNA library."
                    "'unstranded' - reads from the left-most end of the fragment",
                    "(in transcript coordinates) map to the transcript strand, and",
                    "the right-most end maps to the opposite strand.",
                    "'forward' - same as 'unstranded' except we enforce the rule that",
                    "the left-most end of the fragment (in transcript coordinates) is",
                    "the first sequenced (or only sequenced for single-end reads).",
                    "Equivalently, it is assumed that only the strand generated",
                    "during second strand synthesis is sequenced. Used for Ligation and",
                    "Standard SOLiD.",
                    "'reverse' - same as 'unstranded' except we enforce the rule that",
                    "the right-most end of the fragment (in transcript coordinates) is",
                    "the first sequenced (or only sequenced for single-end reads).",
                    "Equivalently, it is assumed that only the strand generated during",
                    "first strand synthesis is sequenced. Used for dUTP, NSR, and NNSR.",
                    "Default: first 'auto' (try to detect strand from the XS tag",
                    "of the read), then downgrade to 'unstranded'"
                ]
            ),
            type=str,
            default="auto",
            choices=["auto", "forward", "reverse", "unstranded"]
        )
        intron_parser.add_argument(
            "--chr",
            help="Select chromosomes to process. Default: only main chromosomes",
            type=str,
            nargs="*",
            default=MAIN_CRH
        )
        intron_parser.add_argument(
            "--savereads",
            help="Export processed reads into the BAM file. Default: False",
            action="store_true"
        )
        self.add_common_arguments(intron_parser)

        # Junction count parameters
        junction_parser = subparsers.add_parser(
            "juncount",
            parents=[parent_parser],
            help="Count junctions reads"
        )
        junction_parser.set_defaults(func=count_junctions)
        junction_parser.add_argument(
            "--bam",
            help="Path to the coordinate-sorted indexed BAM file",
            type=str,
            required=True
        )
        junction_parser.add_argument(
            "--chr",
            help="Select chromosomes to process. Default: only main chromosomes",
            type=str,
            nargs="*",
            default=MAIN_CRH
        )
        junction_parser.add_argument(
            "--savereads",
            help="Export processed reads into the BAM file. Default: False",
            action="store_true"
        )
        self.add_common_arguments(junction_parser)

        # Aggregate parameters
        aggregate_parser = subparsers.add_parser(
            "aggregate",
            parents=[parent_parser],
            help="Aggregate read counts produced by intcount and juncount"
        )
        aggregate_parser.set_defaults(func=aggregate)
        aggregate_parser.add_argument(
            "--juncounts",
            help=" ".join(
                [
                    "Path the junction counts files. Coordinates are treated as 0-based.",
                    "If provided with --intcounts, the number and the order should",
                    "correspond to --intcounts."
                ]
            ),
            type=str,
            nargs="*"
        )
        aggregate_parser.add_argument(
            "--intcounts",
            help=" ".join(
                [
                    "Path the intron counts files. Coordinates are treated as 0-based.",
                    "If provided with --juncounts, the number and the order should",
                    "correspond to --juncounts."
                ]
            ),
            type=str,
            nargs="*"
        )
        aggregate_parser.add_argument(
            "--aliases",
            help=" ".join(
                [
                    "Column names to be used for the loaded counts. The number of provided",
                    "aliases should be equal to the number of --intcounts and/or --juncounts",
                    "files. Default: rootname of --intcounts and/or --intcounts files."
                ]
            ),
            type=str,
            nargs="*"
        )
        aggregate_parser.add_argument(
            "--ref",
            help=" ".join(
                [
                    "Path to the gene model reference file. Coordinates are treated as 1-based.",
                    "Required if --juncounts parameter was provided."
                ]
            ),
            type=str,
            required=False
        )
        aggregate_parser.add_argument(
            "--chr",
            help="Select chromosomes to process. Default: only main chromosomes",
            type=str,
            nargs="*",
            default=MAIN_CRH
        )
        aggregate_parser.add_argument(
            "--bed",
            help="Export annotated 0-based coordinates as BED file. Default: False",
            action="store_true"
        )
        self.add_common_arguments(aggregate_parser)

        return general_parser

    def resolve_path(self, selected=None):
        """
        Resolves path of the "selected" parameters.
        The other parameters remain unchanged
        """
        selected = [] if selected is None else selected
        normalized_args = {}
        for key, value in self.args.__dict__.items():
            if key in selected and value is not None:
                if isinstance(value, list):
                    for v in value:
                        normalized_args.setdefault(key, []).append(
                            pathlib.Path(v).resolve()
                        )
                else:
                    normalized_args[key] = pathlib.Path(value).resolve()
            else:
                normalized_args[key] = value
        self.args = argparse.Namespace(**normalized_args)

    def assert_args(self):
        """
        Should be used to assert and fix parameters.
        Also can be used to set default values for not
        set parameters in case the later ones depend on other
        parameters that should be first parsed by argparser
        """
        self.assert_common_args()
        if self.args.func == count_junctions:
            self.assert_args_for_count_junctions()
        elif self.args.func == count_introns:
            self.assert_args_for_count_introns()
        elif self.args.func == aggregate:
            self.assert_args_for_aggregate()

    def assert_args_for_count_junctions(self):
        pass

    def assert_args_for_count_introns(self):
        self.args.strandness = IntRetCat[self.args.strandness.upper()]
        self.args.ref = get_indexed_references(
            location=self.args.ref,
            tmp_location=self.args.tmp,
            selected_chr=self.args.chr,
            only_introns=True
        )

    def assert_args_for_aggregate(self):
        if self.args.juncounts is None and self.args.intcounts is None:
            logging.error("At least one of the --juncounts or --intcounts inputs should be provided")
            sys.exit(1)
        if self.args.juncounts is not None and self.args.intcounts is not None and len(self.args.juncounts) != len(self.args.intcounts):
            logging.error("Number of the provided files for --juncounts and --intcounts should be equal")
            sys.exit(1)
        if self.args.aliases is not None:
            if self.args.juncounts is not None and len(self.args.aliases) != len(self.args.juncounts):
                logging.error("Number of the provided --aliases and --juncounts files should be equal")
                sys.exit(1)
            if self.args.intcounts is not None and len(self.args.aliases) != len(self.args.intcounts):
                logging.error("Number of the provided --aliases and --intcounts files should be equal")
                sys.exit(1)
        if self.args.aliases is None:
            zipped = zip(self.args.juncounts) if self.args.juncounts is not None else zip(self.args.intcounts)
            zipped = zip(*zip(*zipped), self.args.intcounts if self.args.intcounts is not None else self.args.juncounts)
            self.args.aliases = [os.path.commonprefix([a.stem, b.stem]) for a, b in zipped]
        if self.args.juncounts is not None:
            if self.args.ref is None:
                logging.error("--ref parameter is required when using with --intcounts")
                sys.exit(1)
            self.args.ref = get_indexed_references(
                location=self.args.ref,
                tmp_location=self.args.tmp,
                selected_chr=self.args.chr,
                only_introns=False
            )

    def assert_common_args(self):
        self.args.loglevel = getattr(logging, self.args.loglevel.upper())
        setup_logger(logging.root, self.args.loglevel)
        self.args.tmp.mkdir(parents=True, exist_ok=True)                          # safety measure, shouldn't fail
        self.args.output.parent.mkdir(parents=True, exist_ok=True)                # safety measure, shouldn't fail
        self.args.chr = list(map(ChrConverter, self.args.chr))
        if hasattr(self.args, "bam"):
            if not is_bam_indexed(self.args.bam):
                try:
                    pysam.index(str(self.args.bam))                               # attempts to create bai index (will raise if bam is not sorted)
                except pysam.SamtoolsError:
                    logging.error(f"""Failed to index {self.args.bam}. Exiting""")
                    sys.exit(1)


import pandas as pd
import numpy as np
import scanpy as sc
from classificationFunctions import *
from sklearn import metrics
import matplotlib.pyplot as plt
from sklearn.metrics.pairwise import cosine_similarity
from tqdm import tqdm
import warnings


# find the top 500 highly variable features per cell type
def determine_cell_type_hvg(adata, groups_col='cell_type', train_col='disease', train_id='control', n_top_genes=500):
    adata_train = adata[adata.obs[train_col] == train_id]

    # Initialize an empty dictionary to store highly variable genes for each cell type
    hvg_dict = {}

    # Get unique cell types
    cell_types = adata_train.obs[groups_col].unique()

    # Loop through each cell type
    for cell_type in cell_types:
        print(cell_type)
        # Subset the AnnData object
        adata_subset = adata_train[adata_train.obs[groups_col] == cell_type]

        # Check the number of cells
        if adata_subset.shape[0] < 50:
            warnings.warn(
                f"Cell type '{cell_type}' has less than 50 cells ({adata_subset.shape[0]} cells). Highly variable gene computation may not be reliable.")

        if adata_subset.shape[0] < 10:
            hvg_dict[cell_type] = []
            continue

        # Compute highly variable genes
        sc.pp.highly_variable_genes(adata_subset, flavor='seurat_v3', n_top_genes=n_top_genes, layer='counts',span=0.8)

        # Get the top 500 highly variable genes
        hvg = adata_subset.var[adata_subset.var['highly_variable']].index.tolist()

        # Store the result in the dictionary
        hvg_dict[cell_type] = hvg

    # Convert the dictionary to a DataFrame
    hvg_df = pd.DataFrame.from_dict(hvg_dict, orient='index').transpose()

    adata.uns['top_n_hvg'] = hvg_df

    return adata


def determine_control_cells_cosine_similarity(adata, groups_col='cell_type', train_col='disease', train_id='control', hvg_matrix_field='top_n_hvg', save_obs_key='cosine_similarity',donor_key='donor_ids',disease_key='disease_type'):

    # split into train dataset and test dataset
    train_controls = adata[adata.obs[train_col] == train_id, :]
    test = adata[~(adata.obs[train_col] == train_id), :]

    print("finished train test split")

    adata.obs[(save_obs_key + '_predicted_labels')] = ''
    adata.obs[(save_obs_key + '_decision_values')] = ''
    cell_types = adata.obs[groups_col].unique()

    for cluster in tqdm(cell_types, desc="Predicting cell types with cosine similarity..."):

        print(cluster)
        top_500_var_feats_all_cell_types = adata.uns[hvg_matrix_field]
        var_feats_s = top_500_var_feats_all_cell_types.loc[:, cluster]

        # Remove missing values (NaNs)
        var_feats_s = var_feats_s.dropna()

        if len(var_feats_s) < 50:
            warnings.warn(
                f"Cell type '{cluster}' has less than 50 highly variable genes ({len(var_feats_s)} genes). not performing one-class SVM.")
            continue

        # subset with the top 500 feats and the cells of that cluster
        train_controls_s = train_controls[train_controls.obs[groups_col] == cluster, var_feats_s]
        test_s = test[test.obs[groups_col] == cluster, var_feats_s]

        train_adata, test_adata, predicted_labels, row_mean_similarity = gene_cosine_similarity(train_adata=train_controls_s, test_adata=test_s)
        adata.obs.loc[test_s.obs_names, (save_obs_key + '_predicted_labels')] = predicted_labels
        adata.obs.loc[test_s.obs_names, (save_obs_key + '_similarity_values')] = row_mean_similarity

        test_s.obs.loc[test_s.obs_names, (save_obs_key + '_predicted_labels')] = predicted_labels
        test_s.obs.loc[test_s.obs_names, (save_obs_key + '_similarity_values')] = row_mean_similarity

        plt.close("all")
        plt.figure()
        sc.pl.pca(train_controls_s, color=donor_key, legend_loc='right margin', dimensions=[(0, 1)])
        train_pca_file = f"{cluster.replace(' ', '_')}_train_pca.png"
        plt.savefig(train_pca_file, bbox_inches='tight')
        plt.close("all")

        plt.figure()
        sc.pl.pca(test_s, color=[donor_key, disease_key, (save_obs_key + '_similarity_values')],
                  dimensions=[(0, 1)])
        test_pca_file = f"{cluster.replace(' ', '_')}_test_pca.png"
        plt.savefig(test_pca_file, bbox_inches='tight')
        plt.close()
        print("finished predicting")

    return adata


def determine_control_cells_one_class_svm(adata, groups_col='cell_type', train_col='disease', train_id='control', hvg_matrix_field='top_n_hvg', svm_matrix='pca', save_obs_key='one_class_svm', donor_key='donor_ids',disease_key='disease_type'):

    # split into train dataset and test dataset
    train_controls = adata[adata.obs[train_col] == train_id, :]
    test = adata[~(adata.obs[train_col] == train_id), :]

    print("finished train test split")

    adata.obs[(save_obs_key + '_predicted_labels')] = ''
    adata.obs[(save_obs_key + '_decision_values')] = ''
    cell_types = adata.obs[groups_col].unique()

    for cluster in tqdm(cell_types, desc="Predicting cell types with one-class SVM..."):

        print(cluster)
        top_500_var_feats_all_cell_types = adata.uns[hvg_matrix_field]
        var_feats_s = top_500_var_feats_all_cell_types.loc[:, cluster]


        # Remove missing values (NaNs)
        var_feats_s = var_feats_s.dropna()

        if len(var_feats_s) < 50:
            warnings.warn(
                f"Cell type '{cluster}' has less than 50 highly variable genes ({len(var_feats_s)} genes). not performing one-class SVM.")
            continue

        # subset with the top 500 feats and the cells of that cluster
        train_controls_s = train_controls[train_controls.obs[groups_col] == cluster, var_feats_s]
        test_s = test[test.obs[groups_col] == cluster, var_feats_s]

        # one class svm
        train_adata, test_adata, predicted_labels, decision_values = one_class_svm(train_adata=train_controls_s, test_adata=test_s, nu=0.1, matrix=svm_matrix)

        adata.obs.loc[test_s.obs_names, (save_obs_key + '_predicted_labels')] = predicted_labels
        adata.obs.loc[test_s.obs_names, (save_obs_key + '_decision_values')] = decision_values

        test_s.obs.loc[test_s.obs_names, (save_obs_key + '_predicted_labels')] = predicted_labels
        test_s.obs.loc[test_s.obs_names, (save_obs_key + '_decision_values')] = decision_values

        print("finished predicting")

        plt.close("all")
        plt.figure()
        sc.pl.pca(train_controls_s, color=donor_key, legend_loc='right margin', dimensions=[(0, 1)])
        train_pca_file = f"{cluster.replace(' ', '_')}_train_pca.png"
        plt.savefig(train_pca_file, bbox_inches='tight')
        plt.close("all")

        plt.figure()
        sc.pl.pca(test_s, color=[donor_key, disease_key, (save_obs_key + '_decision_values')],
                  dimensions=[(0, 1)])
        test_pca_file = f"{cluster.replace(' ', '_')}_test_pca.png"
        plt.savefig(test_pca_file, bbox_inches='tight')
        plt.close()

    return adata




import pandas as pd
import numpy as np
from scipy.stats import t


def unique_marker_finder(input_df, groups):

    # convert nmf_clusters to list object
    nmf_clusters_list = groups['cluster'].tolist()

    # ensure fold matrix has the same columns as rows/index of nmf_clusters
    fold_matrix = input_df.loc[groups.index, :]

    corr_df, p_val_df = marker_finder(input_df=fold_matrix, groups=nmf_clusters_list)

    # find the top n markers for each cluster
    top_cluster = corr_df.idxmax(axis=1)  # top cluster for each marker
    top_r = corr_df.max(axis=1)  # top r value for each marker

    # Assign each marker to its top scoring cluster
    markers_df = pd.DataFrame({"marker": top_cluster.index.values,
                               "top_cluster": top_cluster.values,
                               "pearson_r": top_r.values,
                               "p_value": [p_val_df.loc[i, j] for i, j in zip(top_cluster.index.values, top_cluster.values)]})

    markers_df = markers_df.sort_values(by=["top_cluster", "pearson_r"], ascending=[True, False])

    return markers_df


def create_final_marker_heatmap(input_df, markers_df, groups_df):

    # Filter the input_df rows with the index of markers_df
    final_marker_heatmap = input_df.loc[:, markers_df['marker']]

    final_marker_heatmap = final_marker_heatmap.transpose()
    final_marker_heatmap = scale_final_marker_heatmap(final_marker_heatmap)

    # Create a new row 'column_clusters-flat' based on the 'cluster' values from groups_df
    column_clusters = groups_df.loc[final_marker_heatmap.columns].transpose()
    final_marker_heatmap = pd.concat([column_clusters, final_marker_heatmap])
    sorted_columns = final_marker_heatmap.iloc[0].sort_values().index
    final_marker_heatmap = final_marker_heatmap[sorted_columns]

    # Create a new column 'row_clusters-flat' based on the 'top_cluster' values from markers_df
    row_clusters = markers_df.loc[final_marker_heatmap.index[1:], 'top_cluster']
    row_clusters = pd.concat([pd.Series('column_clusters-flat'), row_clusters])
    final_marker_heatmap.insert(0, 'row_clusters-flat', row_clusters)

    final_marker_heatmap.index.values[0] = 'column_clusters-flat'

    return final_marker_heatmap


def scale_final_marker_heatmap(final_marker_heatmap):

    # Drop the first row and column to get the matrix
    mat = final_marker_heatmap.values
    row_names = final_marker_heatmap.index
    col_names = final_marker_heatmap.columns

    # Compute row medians and perform row median centering
    row_medians = np.median(mat, axis=1)
    mat_n = mat - row_medians[:, None]

    # Create a dataframe for seaborn heatmap
    final_marker_heatmap_scaled = pd.DataFrame(mat_n, index=row_names, columns=col_names)

    return final_marker_heatmap_scaled


def marker_finder_wrapper(input_df, groups, top_n=60, rho_threshold=0.2, marker_finder_rho=0.3):

    # get unique markers
    markers_df_og = unique_marker_finder(input_df, groups)
    markers_df_og.index = markers_df_og['marker']
    markers_df = markers_df_og.copy()

    # count number of markers above rho_threshold per cluster
    markers_df = markers_df[markers_df['pearson_r'] >= rho_threshold]

    # clusters with fewer than two markers above the supplied rho threshold are excluded
    markers_df = markers_df.groupby('top_cluster').filter(lambda x: len(x) >= 3)

    # get Top n correlated marker for each cluster
    markers_df = markers_df.groupby('top_cluster').head(top_n)

    # keep only those markers whose pearson r is above rho_threshold
    markers_df = markers_df[markers_df['pearson_r'] >= marker_finder_rho]

    # create an output similar to FinalMarkerHeatmap file
    final_marker_heatmap = create_final_marker_heatmap(input_df, markers_df=markers_df, groups_df=groups)

    return markers_df_og, markers_df, final_marker_heatmap


def marker_finder(input_df, groups):
    """
    Function to find pearson correlation coefficient values and p-values for
    the given data and groups for groups to test. The function will perform a
    Pearson correlation of the input_df feature values to an "idealized"
    group specific expression vector, where each observation in a given group
    is set to a value of 1, and the observations in other groups are set to 0.

    Arguments
    ---------
    input_df : pandas.DataFrame
        Data Frame with observations as index and features as columns (Required)

    groups : list[str]
        List-like of specified groups corresponding to observations from the
        input_df. The order of groups should match the order in input_df.index
        (Required)

    Returns
    -------
    pandas.DataFrame, pandas.DataFrame
        The first item in the tuple is a pandas.DataFrame containing the pearson
        correlation coefficient values for each marker to the idealized vector
        for each cluster.
        The second item is also a pandas.DataFrame, but contains the p-values
        for each comparison.
    """
    ideal_vectors = pd.get_dummies(groups)
    ideal_vectors.index = input_df.index.values
    degrees_f = input_df.shape[0] - 2
    r_df = pearson_corr_df_to_df(input_df, ideal_vectors)
    r_df = r_df.dropna()
    t_df = r_df * np.sqrt(degrees_f) / np.sqrt(1 - (r_df ** 2))
    p_df = t_df.map(lambda x: t.sf(abs(x), df=degrees_f) * 2)
    return r_df, p_df


# Computes pearson correlation coefficient for pairwise columns to columns
# of input DataFrames
def pearson_corr_df_to_df(df1, df2):
    norm1 = df1 - df1.mean(axis=0)
    norm2 = df2 - df2.mean(axis=0)
    sqsum1 = (norm1**2).sum(axis=0)
    sqsum2 = (norm2**2).sum(axis=0)
    return (norm1.T @ norm2) / np.sqrt(sqsum1.apply(lambda x: x*sqsum2))



import pytest
import pathlib
from altanalyze3.utilities.helpers import (
    get_md5_sum,
    get_tmp_suffix
)


DATA_FOLDER = pathlib.Path(__file__).resolve().parents[1].joinpath("data")


@pytest.mark.parametrize(
    "location, control_md5_sum",
    [
        ("gene_model_all.tsv", "2b8531420d7916b9da21102586a96015")
    ]
)
def test_get_md5_sum(location, control_md5_sum):
    calculated_md5_sum = get_md5_sum(DATA_FOLDER.joinpath(location))
    assert calculated_md5_sum == control_md5_sum


@pytest.mark.parametrize(
    "length_without_dot",
    [10, 12, 3]
)
def test_get_tmp_suffix(length_without_dot):
    tmp_suffix = get_tmp_suffix(length_without_dot)
    assert len(tmp_suffix) == length_without_dot + 1


import pandas as pd
import numpy as np
import nimfa
from sklearn.preprocessing import scale
from numpy import linalg as LA


# determine the rank of the NMF decomposition
def determine_nmf_ranks(df):
    # "To estimate the rank of the matrix (i.e. clusters) for SNMF, the ICGS Guide3 matrix is z-score normalized and its eigenvalues are calculated."

    # Convert X to a numpy array
    X = np.array(df)
    g = float(X.shape[0])  # Number of rows/genes
    c = float(X.shape[1])  # Number of columns/pseudobulks
    print(g)
    print(c)

    # Scale the data (zero mean and unit variance)
    X = scale(X)
    Xt = np.transpose(X)  # Transpose of X/df

    # Calculate muTW and sigmaTW for Tracy-Widom distribution
    muTW = (np.sqrt(g - 1) + np.sqrt(c)) ** 2.0
    sigmaTW = (np.sqrt(g - 1) + np.sqrt(c)) * (1.0 / np.sqrt(g - 1) + 1.0 / np.sqrt(c)) ** (1.0 / 3.0)

    # Compute the covariance matrix
    sigmaHat = np.dot(Xt, X)

    # Calculate the threshold boundary
    boundary = 3.273 * sigmaTW + muTW
    print(boundary)

    # Compute the eigenvalues of the covariance matrix
    w, v = LA.eig(sigmaHat)
    w = w.tolist()

    k = 0
    for i in range(len(w)):
        try:
            # Count the number of eigenvalues greater than the boundary
            if w[i] > boundary:
                k += 1
        except Exception:
            if w[i].real > boundary:
                k += 1

    est_k = 2*k   # Estimate the rank of the matrix

    return est_k


# write the pseudocode here ##
'''
Notes:
1. the input for the nmf analysis cannot contain NA values 
2. Doing NMF on a matrix (in this case the imputed PSI matrix) decomposes the input matrix into W (basis matrix of size # of samples by rank/components) and H (coefficient matrix) matrices 
3. we will consider W matrix for all downstream analyses

Main steps in the run_nmf function: 
1. Run NMF and get the W matrix with "rank" parameter indicating the number of clusters NMF should determine. The rank parameter is determined before this analysis. 

'''


def run_nmf(df, rank):
    mat = df.to_numpy()
    mat_t = mat.transpose()

    # sd cannot be 0 across all rows/cols? ##what does nimfa snmf do when some columns or even rows are all 0s? ## see source for their checks
    # enter try statement here
    w = None
    try:
        nmf = nimfa.Snmf(mat_t, seed="nndsvd", rank=int(rank), max_iter=20, n_run=5,
                         track_factor=True)  # n_run was originally 10 to keep the results stable
        nmf_fit = nmf()
        w = nmf_fit.basis()
    except ValueError:
        w = mat  # this needs to change -- how are errors handled?

    finally:
        nmf_matrix = pd.DataFrame(w.transpose(), columns=df.columns)

    nmf_clusters = binarize_nmf(w)
    nmf_clusters.index = df.columns

    return nmf_matrix, nmf_clusters


def binarize_nmf(w):
    w = w.transpose()
    # components = w.shape[0]  # this is number of rows/components
    samples = w.shape[1]  # number of columns/samples

    max_decision_value = np.zeros((samples, 1))

    for i in range(samples):
        # print(i)
        # i-th entry in max_decision_value is the row number with the maximum value in the i-th column of w
        max_decision_value[i] = np.argmax(w[:, i])

    nmf_clusters = pd.DataFrame(data=max_decision_value, columns=['cluster'], dtype='int64')

    return nmf_clusters



import os,sys
import pandas as pd
import numpy as np
from copy import deepcopy
import argparse
import warnings
import aiofiles  # Asynchronous file I/O
import asyncio   # For async handling
from tqdm import tqdm  # For the progress bar

# Chunk size and batch size
CHUNK_SIZE = 50  # Number of genes to process before writing
BATCH_SIZE = 10  # Number of genes per async file write

def initialize_output_file(file_path, header):
    with open(file_path, 'w') as f:
        f.write('\t'.join(header) + '\n')  # Write header once at the beginning

# Original method for clique finding
def find_uid_in_clique(the_uid, region, strand, uid2coords):
    clique = {the_uid: region}
    region_start, region_end = map(int, region)

    for uid, coords in uid2coords.items():
        if uid != the_uid:
            coords_start, coords_end = map(int, coords)
            # Check overlap
            if strand == '+':
                if max(region_start, coords_start) - min(region_end, coords_end) < 0:
                    clique[uid] = coords
            elif strand == '-':
                if max(region_end, coords_end) - min(region_start, coords_start) < 0:
                    clique[uid] = coords

    return clique

# Early exit conditional checking to optimize
def is_valid(mat, uid):
    min_reads = 5
    num_incl_events = np.count_nonzero(mat[0, :] >= min_reads)
    if num_incl_events <= 1:
        return False  # Early exit if inclusion events are too low

    num_excl_events = np.count_nonzero(mat[1:, :].sum(axis=0) >= min_reads)
    if num_excl_events <= 1:
        return False  # Early exit if exclusion events are too low

    total_number_junctions = np.count_nonzero(mat.sum(axis=0) >= min_reads)
    if total_number_junctions <= 2:
        return False  # Early exit if total junction count is too low

    return True

# Core PSI calculation with early exit conditions
def calculate_psi_core(clique, uid, count, sample_columns):
    sub_count = count.loc[list(clique.keys()), sample_columns]
    mat = sub_count.values
    with np.errstate(divide='ignore', invalid='ignore'):
        psi = mat[0, :] / mat.sum(axis=0)

    if sub_count.shape[0] > 1:
        bg_uid = sub_count.index.tolist()[np.argmax(mat[1:, :].sum(axis=1)) + 1]
        cond = is_valid(mat, uid)
    else:
        bg_uid = 'None'
        cond = False

    return psi, bg_uid, cond

# Asynchronously write to file
async def write_to_file(file_path, data, write_header):
    async with aiofiles.open(file_path, 'a') as f:
        header = '\t'.join(data.columns) + '\n' if write_header else ''
        await f.write(header)
        await f.write(data.to_csv(sep='\t', index=False, header=False))

# PSI calculation for a chunk of genes, writing in batches
def calculate_psi_per_gene(count, outdir, write_header=True, result_batch=None):
    return_data = []
    count = count.set_index('uid')
    uid2coords = count.apply(lambda x: [x['start'], x['end']], axis=1, result_type='reduce').to_dict()

    for row in count.iterrows():
        uid = row[0]
        region = uid2coords[uid]
        strand = row[1]['strand']
        clique = find_uid_in_clique(uid, region, strand, uid2coords)
        index_list = deepcopy(row[1].index.tolist())
        for c in ['gene', 'chrom', 'start', 'end', 'strand']:
            index_list.remove(c)
        sample_columns = index_list
        psi_values, bg_uid, cond = calculate_psi_core(clique, uid, count, sample_columns)

        # Only keep PSI results that meet the range condition
        if np.nanmax(psi_values) - np.nanmin(psi_values) >= 0.1 and cond:
            data = (uid+'|'+bg_uid, *psi_values)
            return_data.append(data)

    if return_data:
        df = pd.DataFrame.from_records(return_data, columns=['uid-bg_uid'] + sample_columns)
        result_batch.append(df)


async def main(junction_path=None, query_gene=None, outdir=None):
    import time
    start_time = time.time()

    # Extract the header before initializing the output file
    with open(junction_path, 'r') as f:
        header = ['uid']+next(f).strip().split('\t')  # Read and split header from input file

    # Initialize output file with header
    initialize_output_file(outdir, header[1:])  # Call the function here

    col_uid = []
    col_gene = []
    col_chrom = []
    col_start = []
    col_end = []
    col_strand = []
    data = []

    current_gene = None
    write_header = True
    result_batch = []

    total_lines = sum(1 for line in open(junction_path))  # Count lines for progress bar
    with tqdm(total=total_lines, desc="Processing junctions") as pbar:
        with open(junction_path, 'r') as f:
            header = next(f).strip().split('\t')
            for line in f:
                pbar.update(1)  # Update progress bar
                row = line.strip().split('\t')
                uid, coords = row[0].split('=')
                chrom, coords = coords.split(':')
                start, end = coords.split('-')
                strand = '+' if start < end else '-'
                gene = uid.split(':')[0]

                # If processing a specific gene, skip others
                if query_gene and gene != query_gene:
                    continue

                if current_gene and current_gene != gene:
                    count = pd.DataFrame(data, columns=header[1:])
                    for name, col in zip(['uid', 'gene', 'chrom', 'start', 'end', 'strand'],
                                         [col_uid, col_gene, col_chrom, col_start, col_end, col_strand]):
                        count[name] = col

                    # Process chunk of genes and write to disk
                    calculate_psi_per_gene(count, outdir, write_header, result_batch)
                    if len(result_batch) >= BATCH_SIZE:
                        # Write results in batches
                        await write_to_file(outdir, pd.concat(result_batch), write_header)
                        result_batch.clear()

                    write_header = False

                    col_uid, col_gene, col_chrom, col_start, col_end, col_strand, data = [], [], [], [], [], [], []
                
                current_gene = gene
                col_uid.append(uid)
                col_gene.append(gene)
                col_chrom.append(chrom)
                col_start.append(start)
                col_end.append(end)
                col_strand.append(strand)
                data.append([float(x) if x.replace('.', '', 1).isdigit() else np.nan for x in row[1:]])

        if data:
            count = pd.DataFrame(data, columns=header[1:])
            for name, col in zip(['uid', 'gene', 'chrom', 'start', 'end', 'strand'],
                                 [col_uid, col_gene, col_chrom, col_start, col_end, col_strand]):
                count[name] = col
            calculate_psi_per_gene(count, outdir, write_header, result_batch)
            count={}

        if result_batch:
            await write_to_file(outdir, pd.concat(result_batch), write_header)
            
    end_time = time.time()
    print(f"Total time: {end_time - start_time:.2f} seconds")

if __name__ == '__main__':

    parser = argparse.ArgumentParser(description='Compute PSI from count junction')
    parser.add_argument('--junction', type=str, default=None, help='the path to the junction count')
    parser.add_argument('--gene', type=str, default=None, help='gene you want to compute PSI')
    parser.add_argument('--outdir', type=str, default=None, help='output dir for the output file')
    args = parser.parse_args()

    junction_path = args.junction
    query_gene = args.gene
    outdir = args.outdir

    asyncio.run(main(junction_path=junction_path, query_gene=None, outdir=outdir))
